Index: src/compiler/arm/instruction-selector-arm.cc |
diff --git a/src/compiler/arm/instruction-selector-arm.cc b/src/compiler/arm/instruction-selector-arm.cc |
index af55951f64e51831a4a119dcc1300527f8e374b9..76a0acb410f0b8fcec908285b8935c94df6e8751 100644 |
--- a/src/compiler/arm/instruction-selector-arm.cc |
+++ b/src/compiler/arm/instruction-selector-arm.cc |
@@ -1080,12 +1080,11 @@ void InstructionSelector::VisitFloat64RoundTiesAway(Node* node) { |
} |
-void InstructionSelector::VisitCall(Node* node, BasicBlock* handler, |
- CallMode call_mode) { |
+void InstructionSelector::VisitCall(Node* node, BasicBlock* handler) { |
ArmOperandGenerator g(this); |
const CallDescriptor* descriptor = OpParameter<const CallDescriptor*>(node); |
- FrameStateDescriptor* frame_state_descriptor = NULL; |
+ FrameStateDescriptor* frame_state_descriptor = nullptr; |
if (descriptor->NeedsFrameState()) { |
frame_state_descriptor = |
GetFrameStateDescriptor(node->InputAt(descriptor->InputCount())); |
@@ -1094,12 +1093,11 @@ void InstructionSelector::VisitCall(Node* node, BasicBlock* handler, |
CallBuffer buffer(zone(), descriptor, frame_state_descriptor); |
// Compute InstructionOperands for inputs and outputs. |
// TODO(turbofan): on ARM64 it's probably better to use the code object in a |
// register if there are multiple uses of it. Improve constant pool and the |
// heuristics in the register allocator for where to emit constants. |
InitializeCallBuffer(node, &buffer, true, false); |
- // TODO(dcarney): might be possible to use claim/poke instead |
// Push any stack arguments. |
for (Node* node : base::Reversed(buffer.pushed_nodes)) { |
Emit(kArmPush, g.NoOutput(), g.UseRegister(node)); |
@@ -1107,21 +1105,20 @@ void InstructionSelector::VisitCall(Node* node, BasicBlock* handler, |
// Pass label of exception handler block. |
CallDescriptor::Flags flags = descriptor->flags(); |
- if (handler != nullptr) { |
+ if (handler) { |
flags |= CallDescriptor::kHasExceptionHandler; |
buffer.instruction_args.push_back(g.Label(handler)); |
} |
// Select the appropriate opcode based on the call type. |
- bool is_tail_call = call_mode == TAIL_CALL; |
InstructionCode opcode; |
switch (descriptor->kind()) { |
case CallDescriptor::kCallCodeObject: { |
- opcode = is_tail_call ? kArchTailCallCodeObject : kArchCallCodeObject; |
+ opcode = kArchCallCodeObject; |
break; |
} |
case CallDescriptor::kCallJSFunction: |
- opcode = is_tail_call ? kArchTailCallJSFunction : kArchCallJSFunction; |
+ opcode = kArchCallJSFunction; |
break; |
default: |
UNREACHABLE(); |
@@ -1130,13 +1127,95 @@ void InstructionSelector::VisitCall(Node* node, BasicBlock* handler, |
opcode |= MiscField::encode(flags); |
// Emit the call instruction. |
- size_t size = is_tail_call ? 0 : buffer.outputs.size(); |
- InstructionOperand* first_output = |
- size > 0 ? &buffer.outputs.front() : nullptr; |
- Instruction* call_instr = |
- Emit(opcode, size, first_output, buffer.instruction_args.size(), |
- &buffer.instruction_args.front()); |
- call_instr->MarkAsCall(); |
+ size_t const output_count = buffer.outputs.size(); |
+ auto* outputs = output_count ? &buffer.outputs.front() : nullptr; |
+ Emit(opcode, output_count, outputs, buffer.instruction_args.size(), |
+ &buffer.instruction_args.front())->MarkAsCall(); |
+} |
+ |
+ |
+void InstructionSelector::VisitTailCall(Node* node) { |
+ ArmOperandGenerator g(this); |
+ CallDescriptor const* descriptor = OpParameter<CallDescriptor const*>(node); |
+ DCHECK_NE(0, descriptor->flags() & CallDescriptor::kSupportsTailCalls); |
+ DCHECK_EQ(0, descriptor->flags() & CallDescriptor::kPatchableCallSite); |
+ DCHECK_EQ(0, descriptor->flags() & CallDescriptor::kNeedsNopAfterCall); |
+ |
+ // TODO(turbofan): Relax restriction for stack parameters. |
+ if (descriptor->UsesOnlyRegisters() && |
+ descriptor->HasSameReturnLocationsAs( |
+ linkage()->GetIncomingDescriptor())) { |
+ CallBuffer buffer(zone(), descriptor, nullptr); |
+ |
+ // Compute InstructionOperands for inputs and outputs. |
+ // TODO(turbofan): on ARM it's probably better to use the code object in a |
+ // register if there are multiple uses of it. Improve constant pool and the |
+ // heuristics in the register allocator for where to emit constants. |
+ InitializeCallBuffer(node, &buffer, true, false); |
+ |
+ DCHECK_EQ(0u, buffer.pushed_nodes.size()); |
+ |
+ // Select the appropriate opcode based on the call type. |
+ InstructionCode opcode; |
+ switch (descriptor->kind()) { |
+ case CallDescriptor::kCallCodeObject: |
+ opcode = kArchTailCallCodeObject; |
+ break; |
+ case CallDescriptor::kCallJSFunction: |
+ opcode = kArchTailCallJSFunction; |
+ break; |
+ default: |
+ UNREACHABLE(); |
+ return; |
+ } |
+ opcode |= MiscField::encode(descriptor->flags()); |
+ |
+ // Emit the tailcall instruction. |
+ Emit(opcode, 0, nullptr, buffer.instruction_args.size(), |
+ &buffer.instruction_args.front()); |
+ } else { |
+ FrameStateDescriptor* frame_state_descriptor = |
Jarin
2015/05/05 09:15:24
There seems to be some serious overlap with the Vi
Benedikt Meurer
2015/05/05 09:40:21
As said, we will unify VisitCall and VisitTailCall
|
+ descriptor->NeedsFrameState() |
+ ? GetFrameStateDescriptor( |
+ node->InputAt(static_cast<int>(descriptor->InputCount()))) |
+ : nullptr; |
+ |
+ CallBuffer buffer(zone(), descriptor, frame_state_descriptor); |
+ |
+ // Compute InstructionOperands for inputs and outputs. |
+ // TODO(turbofan): on ARM it's probably better to use the code object in a |
+ // register if there are multiple uses of it. Improve constant pool and the |
+ // heuristics in the register allocator for where to emit constants. |
+ InitializeCallBuffer(node, &buffer, true, false); |
+ |
+ // Push any stack arguments. |
+ for (Node* node : base::Reversed(buffer.pushed_nodes)) { |
+ Emit(kArmPush, g.NoOutput(), g.UseRegister(node)); |
+ } |
+ |
+ // Select the appropriate opcode based on the call type. |
+ InstructionCode opcode; |
+ switch (descriptor->kind()) { |
+ case CallDescriptor::kCallCodeObject: { |
+ opcode = kArchCallCodeObject; |
+ break; |
+ } |
+ case CallDescriptor::kCallJSFunction: |
+ opcode = kArchCallJSFunction; |
+ break; |
+ default: |
+ UNREACHABLE(); |
+ return; |
+ } |
+ opcode |= MiscField::encode(descriptor->flags()); |
+ |
+ // Emit the call instruction. |
+ size_t const output_count = buffer.outputs.size(); |
+ auto* outputs = output_count ? &buffer.outputs.front() : nullptr; |
+ Emit(opcode, output_count, outputs, buffer.instruction_args.size(), |
+ &buffer.instruction_args.front())->MarkAsCall(); |
+ Emit(kArchRet, 0, nullptr, output_count, outputs); |
+ } |
} |
@@ -1094,12 +1093,11 @@ void InstructionSelector::VisitCall(Node* node, BasicBlock* handler, |
CallBuffer buffer(zone(), descriptor, frame_state_descriptor); |
// Compute InstructionOperands for inputs and outputs. |
- // TODO(turbofan): on ARM64 it's probably better to use the code object in a |
+ // TODO(turbofan): on ARM it's probably better to use the code object in a |
// register if there are multiple uses of it. Improve constant pool and the |
// heuristics in the register allocator for where to emit constants. |